
# THIS FILE IS GENERATED USING utils.py
# IF YOU'RE ADDING/REMOVING SOME SHADER, DO NOT EDIT THIS FILE DIRECTLY
# INSTEAD DO WHATEVER YOU WANT WITH THE SHADERS AND RUN utils.py

from .base import CustomTransition
from kivy.properties import StringProperty

available_transitions = ["Angular","Bounce","BowTieHorizontal","BowTieVertical","Burn","ButterflyWaveScrawler","CannabisLeaf","Circle","CircleOpen","ColorPhase","ColourDistance","CrazyParametricFun","CrossHatch","CrossWarp","CrossZoom","Cube","DirectionalEasing","DirectionalWarp","DirectionalWipe","Displacement","DoomScreenTransition","Doorway","Dreamy","DreamyZoom","Fade","FadeColor","FadeGrayscale","FilmBurn","FlyEye","GlitchDisplace","GlitchMemories","GridFlip","Heart","Hexagonalize","Kaleidoscope","LeftRight","LinearBlur","Luma","LuminanceMelt","Morph","Mosaic","MultiplyBlend","PageCurl","Perlin","Pinwheel","Pixelize","PolarFunction","PolkaDotsCurtain","Radial","RandomNoise","RandomSquares","Ripple","RotateScaleFade","RotateTransition","SimpleZoom","SquaresWire","Squeeze","StereoViewer","Swap","Swirl","TangentMotionBlur","TopBottom","UndulatingBurnOut","WaterDrop","Wind","WindowBlinds","WindowSlice","WipeDown","WipeLeft","WipeRight","WipeUp","ZoomInCircles",]
__all__ = available_transitions

class Angular(CustomTransition):
    glsl_file = StringProperty("Angular.glsl")



class Bounce(CustomTransition):
    glsl_file = StringProperty("Bounce.glsl")



class BowTieHorizontal(CustomTransition):
    glsl_file = StringProperty("BowTieHorizontal.glsl")



class BowTieVertical(CustomTransition):
    glsl_file = StringProperty("BowTieVertical.glsl")



class Burn(CustomTransition):
    glsl_file = StringProperty("Burn.glsl")



class ButterflyWaveScrawler(CustomTransition):
    glsl_file = StringProperty("ButterflyWaveScrawler.glsl")



class CannabisLeaf(CustomTransition):
    glsl_file = StringProperty("CannabisLeaf.glsl")



class Circle(CustomTransition):
    glsl_file = StringProperty("Circle.glsl")



class CircleOpen(CustomTransition):
    glsl_file = StringProperty("CircleOpen.glsl")



class ColorPhase(CustomTransition):
    glsl_file = StringProperty("ColorPhase.glsl")



class ColourDistance(CustomTransition):
    glsl_file = StringProperty("ColourDistance.glsl")



class CrazyParametricFun(CustomTransition):
    glsl_file = StringProperty("CrazyParametricFun.glsl")



class CrossHatch(CustomTransition):
    glsl_file = StringProperty("CrossHatch.glsl")



class CrossWarp(CustomTransition):
    glsl_file = StringProperty("CrossWarp.glsl")



class CrossZoom(CustomTransition):
    glsl_file = StringProperty("CrossZoom.glsl")



class Cube(CustomTransition):
    glsl_file = StringProperty("Cube.glsl")



class DirectionalEasing(CustomTransition):
    glsl_file = StringProperty("DirectionalEasing.glsl")



class DirectionalWarp(CustomTransition):
    glsl_file = StringProperty("DirectionalWarp.glsl")



class DirectionalWipe(CustomTransition):
    glsl_file = StringProperty("DirectionalWipe.glsl")



class Displacement(CustomTransition):
    glsl_file = StringProperty("Displacement.glsl")



class DoomScreenTransition(CustomTransition):
    glsl_file = StringProperty("DoomScreenTransition.glsl")



class Doorway(CustomTransition):
    glsl_file = StringProperty("Doorway.glsl")



class Dreamy(CustomTransition):
    glsl_file = StringProperty("Dreamy.glsl")



class DreamyZoom(CustomTransition):
    glsl_file = StringProperty("DreamyZoom.glsl")



class Fade(CustomTransition):
    glsl_file = StringProperty("Fade.glsl")



class FadeColor(CustomTransition):
    glsl_file = StringProperty("FadeColor.glsl")



class FadeGrayscale(CustomTransition):
    glsl_file = StringProperty("FadeGrayscale.glsl")



class FilmBurn(CustomTransition):
    glsl_file = StringProperty("FilmBurn.glsl")



class FlyEye(CustomTransition):
    glsl_file = StringProperty("FlyEye.glsl")



class GlitchDisplace(CustomTransition):
    glsl_file = StringProperty("GlitchDisplace.glsl")



class GlitchMemories(CustomTransition):
    glsl_file = StringProperty("GlitchMemories.glsl")



class GridFlip(CustomTransition):
    glsl_file = StringProperty("GridFlip.glsl")



class Heart(CustomTransition):
    glsl_file = StringProperty("Heart.glsl")



class Hexagonalize(CustomTransition):
    glsl_file = StringProperty("Hexagonalize.glsl")



class Kaleidoscope(CustomTransition):
    glsl_file = StringProperty("Kaleidoscope.glsl")



class LeftRight(CustomTransition):
    glsl_file = StringProperty("LeftRight.glsl")



class LinearBlur(CustomTransition):
    glsl_file = StringProperty("LinearBlur.glsl")



class Luma(CustomTransition):
    glsl_file = StringProperty("Luma.glsl")



class LuminanceMelt(CustomTransition):
    glsl_file = StringProperty("LuminanceMelt.glsl")



class Morph(CustomTransition):
    glsl_file = StringProperty("Morph.glsl")



class Mosaic(CustomTransition):
    glsl_file = StringProperty("Mosaic.glsl")



class MultiplyBlend(CustomTransition):
    glsl_file = StringProperty("MultiplyBlend.glsl")



class PageCurl(CustomTransition):
    glsl_file = StringProperty("PageCurl.glsl")



class Perlin(CustomTransition):
    glsl_file = StringProperty("Perlin.glsl")



class Pinwheel(CustomTransition):
    glsl_file = StringProperty("Pinwheel.glsl")



class Pixelize(CustomTransition):
    glsl_file = StringProperty("Pixelize.glsl")



class PolarFunction(CustomTransition):
    glsl_file = StringProperty("PolarFunction.glsl")



class PolkaDotsCurtain(CustomTransition):
    glsl_file = StringProperty("PolkaDotsCurtain.glsl")



class Radial(CustomTransition):
    glsl_file = StringProperty("Radial.glsl")



class RandomNoise(CustomTransition):
    glsl_file = StringProperty("RandomNoise.glsl")



class RandomSquares(CustomTransition):
    glsl_file = StringProperty("RandomSquares.glsl")



class Ripple(CustomTransition):
    glsl_file = StringProperty("Ripple.glsl")



class RotateScaleFade(CustomTransition):
    glsl_file = StringProperty("RotateScaleFade.glsl")



class RotateTransition(CustomTransition):
    glsl_file = StringProperty("RotateTransition.glsl")



class SimpleZoom(CustomTransition):
    glsl_file = StringProperty("SimpleZoom.glsl")



class SquaresWire(CustomTransition):
    glsl_file = StringProperty("SquaresWire.glsl")



class Squeeze(CustomTransition):
    glsl_file = StringProperty("Squeeze.glsl")



class StereoViewer(CustomTransition):
    glsl_file = StringProperty("StereoViewer.glsl")



class Swap(CustomTransition):
    glsl_file = StringProperty("Swap.glsl")



class Swirl(CustomTransition):
    glsl_file = StringProperty("Swirl.glsl")



class TangentMotionBlur(CustomTransition):
    glsl_file = StringProperty("TangentMotionBlur.glsl")



class TopBottom(CustomTransition):
    glsl_file = StringProperty("TopBottom.glsl")



class UndulatingBurnOut(CustomTransition):
    glsl_file = StringProperty("UndulatingBurnOut.glsl")



class WaterDrop(CustomTransition):
    glsl_file = StringProperty("WaterDrop.glsl")



class Wind(CustomTransition):
    glsl_file = StringProperty("Wind.glsl")



class WindowBlinds(CustomTransition):
    glsl_file = StringProperty("WindowBlinds.glsl")



class WindowSlice(CustomTransition):
    glsl_file = StringProperty("WindowSlice.glsl")



class WipeDown(CustomTransition):
    glsl_file = StringProperty("WipeDown.glsl")



class WipeLeft(CustomTransition):
    glsl_file = StringProperty("WipeLeft.glsl")



class WipeRight(CustomTransition):
    glsl_file = StringProperty("WipeRight.glsl")



class WipeUp(CustomTransition):
    glsl_file = StringProperty("WipeUp.glsl")



class ZoomInCircles(CustomTransition):
    glsl_file = StringProperty("ZoomInCircles.glsl")
