from kivy.utils import get_color_from_hex


PALLET_KEY_COLORS = [
	"primary_color",
	"secondary_color",
	"accent_color",
	"background_color",
	"foreground_color",
	"border_color",
	"text_color",
	"heading_color",
	"subheading_color",
	"error_color",
	"success_color",
	"warning_color",
	"info_color",
	"disabled_color",
	"active_color",
	"inactive_color",
	"hover_color",
	"focus_color",
	"selected_color",
	"unselected_color"
]


class Light:
	primary_color = get_color_from_hex("202529")
	secondary_color = get_color_from_hex("fbfbfb")
	accent_color = get_color_from_hex("1fadd2")
	background_color = get_color_from_hex("bf1d15")
	foreground_color = get_color_from_hex("97a1a6")
	border_color = get_color_from_hex("c49a2f")
	text_color = get_color_from_hex("c7cbcb")
	heading_color = get_color_from_hex("f6d49e")
	subheading_color = get_color_from_hex("cfa599")
	error_color = get_color_from_hex("083578")
	success_color = get_color_from_hex("f5f9fa")
	warning_color = get_color_from_hex("16191b")
	info_color = get_color_from_hex("71bca4")
	disabled_color = get_color_from_hex("727a75")
	active_color = get_color_from_hex("127fcf")
	inactive_color = get_color_from_hex("d4bd46")
	hover_color = get_color_from_hex("d1492a")
	focus_color = get_color_from_hex("b8b1ab")
	selected_color = get_color_from_hex("202529")
	unselected_color = get_color_from_hex("fbfbfb")


class Dark:
	primary_color = get_color_from_hex("083578")
	secondary_color = get_color_from_hex("f5f9fa")
	accent_color = get_color_from_hex("16191b")
	background_color = get_color_from_hex("71bca4")
	foreground_color = get_color_from_hex("727a75")
	border_color = get_color_from_hex("127fcf")
	text_color = get_color_from_hex("d4bd46")
	heading_color = get_color_from_hex("d1492a")
	subheading_color = get_color_from_hex("b8b1ab")
	error_color = get_color_from_hex("202529")
	success_color = get_color_from_hex("fbfbfb")
	warning_color = get_color_from_hex("1fadd2")
	info_color = get_color_from_hex("bf1d15")
	disabled_color = get_color_from_hex("97a1a6")
	active_color = get_color_from_hex("c49a2f")
	inactive_color = get_color_from_hex("c7cbcb")
	hover_color = get_color_from_hex("f6d49e")
	focus_color = get_color_from_hex("cfa599")
	selected_color = get_color_from_hex("083578")
	unselected_color = get_color_from_hex("f5f9fa")

